/*
 * Copyright © 2009 Adel Gadllah
 * Copyright © 2009 Red Hat, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *
 */

/**
 * Read /etc/sysconfig/keyboard and convert the contents into an XKB
 * description. Save this configuration as an xorg.conf.d snippet.
 */


#include <glib.h>
#include <glib/gstdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/inotify.h>

#define KBDCONFIG "/etc/sysconfig/keyboard"
#define KEY_OPTIONS "input.xkb.options"
#define CONFIG_DIR "/etc/X11/xorg.conf.d"
#define CONFIG_FILE "00-system-setup-keyboard.conf"
#define CONFIG_PATH CONFIG_DIR "/" CONFIG_FILE
#define TERMINATE_OPTION "terminate:ctrl_alt_bksp,"

static const char *config_snippet =
"# This file is autogenerated by system-setup-keyboard. Any \n" \
"# modifications will be lost.\n\n"			\
"Section \"InputClass\"\n"				\
"	Identifier	\"system-setup-keyboard\"\n"	\
"	MatchIsKeyboard	\"on\"\n"			\
"%s	Option		\"XkbModel\"	\"%s\"\n"	\
"%s	Option		\"XkbLayout\"	\"%s\"\n"	\
"%s	Option		\"XkbVariant\"	\"%s\"\n"	\
"%s	Option		\"XkbOptions\"	\"%s\"\n"	\
"EndSection\n";

static gchar* remove_quotes(gchar *str) {
	gchar *tmp;
	gboolean s_offset = FALSE;

	if(str == NULL)
		return NULL;

	if(str[0] == '"') {
		str++;
		s_offset = TRUE;
	}
	if(str[strlen(str) - 1] == '"') 
		str[strlen(str) - 1] = 0;

	tmp = g_strdup(str);

	if(s_offset)
		str--;
	g_free(str);
	return tmp;	
}

static int update_config(char *executable) {
	GKeyFile *cfg_file;
	gchar *buffer, *conf;

	gchar *model, *layout, *variant, *options;
	gchar *merged_options;
	GHashTable* kbd_models;
	gchar *keytable;
	gchar *property;
	FILE *file;
	struct stat st_sysconfig, st_xorgconf, st_self;

	/* Check if the directory even exists. If not, we have
	 * system-setup-keyboard installed but not the X server. Fair
	 * enough, let's go have a beer. */
	if (g_access(CONFIG_DIR, W_OK | X_OK))
		return 1;

	if (g_access(KBDCONFIG, R_OK))
		return 1;

	if (g_stat(CONFIG_PATH, &st_xorgconf) == 0) {
		/* If our executable has changed since the xorg.conf.d
		 * entry, always run. Otherwise, only run if sysconfig is
		 * newer than xorg.conf.d */
		if (executable == NULL ||
		    !(g_stat(executable, &st_self) == 0 && st_xorgconf.st_mtime < st_self.st_mtime)) {
			if (g_stat(KBDCONFIG, &st_sysconfig) == 0 &&
			    st_sysconfig.st_mtime < st_xorgconf.st_mtime)
			return 0;
		}
	}

	file = g_fopen(CONFIG_PATH, "w");
	if (!file)
		return 1;

	/* Parse the config file */
	cfg_file = g_key_file_new();
	g_file_get_contents(KBDCONFIG, &buffer, NULL, NULL);
	conf = g_strdup_printf("[kbd]\n%s", buffer);
	g_free(buffer);
	g_key_file_load_from_data(cfg_file, conf, strlen(conf), G_KEY_FILE_NONE, NULL);
	property = remove_quotes(g_key_file_get_value(cfg_file, "kbd", "KEYTABLE", NULL));
	if(property == NULL) 
		return 1;

	/* Generate model database */
	kbd_models = g_hash_table_new(g_str_hash, g_str_equal);
#include "keyboards.h"

	/* Read and apply user config */
	keytable = g_hash_table_lookup(kbd_models, (gpointer) property);
	g_free(property);

	model = remove_quotes(g_key_file_get_value(cfg_file, "kbd", "MODEL", NULL));
	layout = remove_quotes(g_key_file_get_value(cfg_file, "kbd", "LAYOUT", NULL));
	variant = remove_quotes(g_key_file_get_value(cfg_file, "kbd", "VARIANT", NULL));
	options = remove_quotes(g_key_file_get_value(cfg_file, "kbd", "OPTIONS", NULL));

	merged_options = g_strdup_printf("%s%s", TERMINATE_OPTION, options ? options : "");
	g_free(options);
	options = merged_options;

	/* if a value is not present, comment out the line.*/
	g_fprintf(file, config_snippet, model ? "" : "#", model,
					layout ? "" : "#", layout,
					variant ? "" : "#", variant,
					options ? "" : "#", options);
	fclose(file);

	/* cleanup */
	g_free(model);
	g_free(layout);
	g_free(variant);
	g_free(options);
	g_free(conf);
	g_hash_table_destroy(kbd_models);

	return 0;
}

static int wait_for_updates(void) {
	int fd, watch;
	struct inotify_event ev;

	fd = inotify_init();

	if (fd == -1)
		return 1;

	watch = inotify_add_watch(fd, KBDCONFIG, IN_MODIFY);

	if (watch == -1)
		return 1;

	while(read(fd, &ev, sizeof(ev)) == sizeof(ev)) {
		if (ev.wd != watch)
			continue;

		if (ev.mask & IN_IGNORED) { /* file deleted? */
			watch = inotify_add_watch(fd, KBDCONFIG, IN_MODIFY);
			if (watch == -1)
				return 1;
		} else if (!(ev.mask & IN_MODIFY))
			continue;

		update_config(NULL);
	}

	inotify_rm_watch(fd, watch);
	close(fd);

	return 0;
}


static void usage(void) {
	printf("system-setup-keyboard is a keyboard layout daemon that monitors "
		"/etc/sysconfig/keyboard and writes out an xorg.conf.d snippet. "
		"For more info see the system-config-keyboard man page.\n");
	printf("Usage: system-setup-keyboard\n");
}

int main(int argc, char **argv) {
	int rc;

	if (argc > 1) {
		usage();
		return 1;
	}

	rc = update_config(argv[0]);
	if (rc != 0)
		return rc;
	rc = wait_for_updates();

	return rc;
}
